import pytest


def test_sles_repos(
    get_instance_repos,
    get_release_value,
    get_sles_repos,
    get_baseproduct,
    determine_architecture
):
    version = [get_release_value('VERSION'), determine_architecture()]

    product = get_baseproduct()
    if 'sap' in product.lower():
        version.append('SAP')
    if 'hpc' in product.lower():
        version.append('HPC')

    version = '-'.join(version)

    missing_repos = []
    repos = get_sles_repos(version)

    if not repos:
        pytest.fail(
            'No repos found for version {0}'.format(version)
        )

    instance_repos = get_instance_repos()

    for repo in repos:
        if repo not in instance_repos:
            missing_repos.append(repo)

    if missing_repos:
        pytest.fail(
            'Missing Repos: \n{0}'.format('\n'.join(missing_repos))
        )



# -*- coding: utf-8 -*-

"""Utility functions."""

# Copyright (c) 2020 SUSE LLC. All rights reserved.
#
# This file is part of img_proof. img_proof provides an api and command line
# utilities for testing images in the Public Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import fnmatch
import logging
import os
import random
import sys
import time

import paramiko
import yaml

from binascii import hexlify
from contextlib import contextmanager
from string import ascii_lowercase
from tempfile import NamedTemporaryFile

from img_proof.ipa_constants import SYNC_POINTS
from img_proof.ipa_exceptions import IpaSSHException, IpaUtilsException

CLIENT_CACHE = {}


def clear_cache(ip=None):
    """Clear the client cache or remove key matching the given ip."""
    if ip:
        with ignored(Exception):
            client = CLIENT_CACHE[ip]
            del CLIENT_CACHE[ip]
            client.close()
    else:
        for client in CLIENT_CACHE.values():
            with ignored(Exception):
                client.close()
        CLIENT_CACHE.clear()


def establish_ssh_connection(ip,
                             ssh_private_key_file,
                             ssh_user,
                             port,
                             attempts=5,
                             timeout=None):
    """
    Establish ssh connection and return paramiko client.

    Raises:
        IpaSSHException: If connection cannot be established
            in given number of attempts.
    """
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    while attempts:
        try:
            client.connect(
                ip,
                port=port,
                username=ssh_user,
                key_filename=ssh_private_key_file,
                timeout=timeout
            )
        except:  # noqa: E722
            attempts -= 1
            time.sleep(10)
        else:
            return client

    raise IpaSSHException(
        'Failed to establish SSH connection to instance.'
    )


def execute_ssh_command(client, cmd):
    """
    Execute given command using paramiko.

    Returns:
        String output of cmd execution.
    Raises:
        IpaSSHException: If stderr returns a non-empty string.
    """
    try:
        stdin, stdout, stderr = client.exec_command(cmd)
        err = stderr.read()
        out = stdout.read()
        if err:
            raise IpaSSHException(out.decode() + err.decode())
    except:  # noqa: E722
        raise
    return out.decode()


def expand_test_files(test_dirs, names):
    """
    Expand the list of test files and test descriptions.

    The list is split on sync points and duplicates in
    each set are removed.

    Returns:
        List of test files split into sets by sync points.
    Raises:
        IpaUtilsException: If names is not a list.
    """
    if not isinstance(names, list):
        raise IpaUtilsException(
            'Names must be a list containing test names'
            ' and/or test descriptions.'
        )

    tests, descriptions = get_test_files(test_dirs)

    expanded_names = []
    for name in names:
        if name in descriptions:
            expanded_names += get_tests_from_description(
                name,
                descriptions
            )
        else:
            expanded_names.append(name)

    return parse_sync_points(expanded_names, tests)


def extract_archive(client, archive_path, extract_path=None):
    """
    Extract the archive in current path using the provided client.

    If extract_path is provided extract the archive there.
    """
    command = 'tar -xf {path}'.format(path=archive_path)

    if extract_path:
        command += ' -C {extract_path}'.format(extract_path=extract_path)

    out = execute_ssh_command(client, command)
    return out


def find_test_file(name, tests):
    """
    Find test file by name, given a list of tests.

    If a specific test case is appended to test
    name, split the case and append to path.

    Raises:
        IpaUtilsException: If test file not found.
    """
    try:
        test_name, test_case = name.split('::', 1)
    except ValueError:
        test_name, test_case = name, None

    path = tests.get(test_name, None)
    if not path:
        raise IpaUtilsException(
            'Test file with name: %s cannot be found.' % test_name
        )

    if test_case:
        path = ''.join([path, '::', test_case])
    return path


def generate_instance_name(name):
    """Generate a new random name for instance."""
    return '%s-%s' % (name, get_random_string(length=5))


def get_public_ssh_key(ssh_private_key_file):
    """Get SSH public key from private key file."""
    pub_key = ssh_private_key_file + '.pub'

    try:
        with open(pub_key, "rb") as key_file:
            key = key_file.read()
    except FileNotFoundError:
        raise IpaUtilsException(
            'SSH public key file: {key_path} cannot be found.'.format(
                key_path=pub_key
            )
        )

    return key


def get_config_values(config_path, section, default='default'):
    """
    Parse ini config file and return a dict of values.

    The provided section overrides any values in default section.
    """
    values = {}

    if not os.path.isfile(config_path):
        raise IpaUtilsException(
            'Config file not found: %s' % config_path
        )

    config = configparser.ConfigParser()

    try:
        config.read(config_path)
    except Exception:
        raise IpaUtilsException(
            'Config file format invalid.'
        )

    try:
        values.update(config.items(default))
    except Exception:
        pass

    try:
        values.update(config.items(section))
    except Exception:
        pass

    return values


def get_host_key_fingerprint(client):
    """Get host key fingerprint of SSH client."""
    return hexlify(
        client.get_transport().get_remote_server_key().get_fingerprint()
    )


def get_random_string(length=12):
    """Create random string of length with ascii lowercase chars."""
    return ''.join(random.choice(ascii_lowercase) for index in range(length))


def get_ssh_client(ip,
                   ssh_private_key_file,
                   ssh_user='root',
                   port=22,
                   timeout=600,
                   wait_period=10):
    """Attempt to establish and test ssh connection."""
    if CLIENT_CACHE.get(ip):
        try:
            execute_ssh_command(CLIENT_CACHE[ip], 'ls')
        except Exception:
            clear_cache(ip)
        else:
            return CLIENT_CACHE[ip]

    start = time.time()
    end = start + timeout

    client = None
    while time.time() < end:
        try:
            client = establish_ssh_connection(
                ip,
                ssh_private_key_file,
                ssh_user,
                port,
                timeout=wait_period
            )
            execute_ssh_command(client, 'ls')
        except:  # noqa: E722
            if client:
                client.close()
            wait_period += wait_period
        else:
            CLIENT_CACHE[ip] = client
            return client

    raise IpaSSHException(
        'Attempt to establish SSH connection failed.'
    )


def get_test_files(test_dirs):
    """
    Walk all test dirs and find all tests and test descriptions.

    Returns:
        A tuple containing a dict mapping test names to full path
        and a dict mapping test descriptions to full path.
    Raises:
        IpaUtilsException: If there are multiple test files or
            test descriptions with the same name. Or, if there is a
            name overlap with a test file and test description.
    """
    tests = {}
    descriptions = {}
    for test_dir in test_dirs:
        if not os.path.exists(test_dir):
            continue

        for root, dirs, files in os.walk(test_dir):
            test_files = fnmatch.filter(files, 'test_*.py')
            description_files = fnmatch.filter(files, 'test_*.yaml')

            for test_file in test_files:
                path = os.path.join(root, test_file)
                name, ext = test_file.split('.')
                if name not in tests:
                    tests[name] = path
                else:
                    raise IpaUtilsException(
                        'Duplicate test file name found: %s, %s'
                        % (path, tests.get(name))
                    )

            for description_file in description_files:
                path = os.path.join(root, description_file)
                name, ext = description_file.split('.')
                if name in tests:
                    raise IpaUtilsException(
                        'Test description name matches test file: %s, %s'
                        % (path, tests.get(name))
                    )
                elif name not in descriptions:
                    descriptions[name] = path
                else:
                    raise IpaUtilsException(
                        'Duplicate test description file name found: %s, %s'
                        % (path, descriptions.get(name))
                    )

    return tests, descriptions


def get_tests_from_description(name,
                               descriptions,
                               parsed=None):
    """
    Recursively collect all tests in test description.

    Args:
        name (str): Yaml test description file name.
        descriptions (dict): Dict of test description name
                             (key) and absolute file paths
                             (value).
        parsed (list): List of description paths which have
                       already been parsed to prevent infinte
                       recursion.
    Returns:
        A list of expanded test files.
    """
    tests = []
    if not parsed:
        parsed = []

    description = descriptions.get(name, None)
    if not description:
        raise IpaUtilsException(
            'Test description file with name: %s cannot be located.'
            % name
        )

    if description in parsed:
        return tests

    parsed.append(description)
    test_data = get_yaml_config(description)

    if 'tests' in test_data:
        tests += test_data.get('tests')

    if 'include' in test_data:
        for description_name in test_data.get('include'):
            tests += get_tests_from_description(
                description_name,
                descriptions,
                parsed
            )

    return tests


def get_yaml_config(config_path):
    """
    Load yaml config file and return dictionary.

    Todo:
        * This will need refactoring similar to the test search.
    """
    config_path = os.path.expanduser(config_path)
    if not os.path.isfile(config_path):
        raise IpaUtilsException(
            'Config file not found: %s' % config_path
        )

    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    return config


@contextmanager
def ignored(*exceptions):
    """Ignore the provided exception(s)."""
    try:
        yield
    except exceptions:
        pass


def parse_sync_points(names, tests):
    """
    Slice list of test names on sync points.

    If test is test file find full path to file.

    Returns:
        A list of test file sets and sync point strings.
    Examples:
        ['test_hard_reboot']
        [set('test1', 'test2')]
        [set('test1', 'test2'), 'test_soft_reboot']
        [set('test1', 'test2'), 'test_soft_reboot', set('test3')]
    """
    test_files = []
    section = set()

    for name in names:
        if name in SYNC_POINTS:
            if section:
                test_files.append(section)

            test_files.append(name)
            section = set()
        else:
            section.add(find_test_file(name, tests))

    if section:
        test_files.append(section)
    return test_files


def parse_test_name(name):
    """Parse and return formatted pytest test name string."""
    test_class = None
    if '::' in name:
        try:
            path, test_class, parens, test_case = name.split('::')
        except ValueError:
            path, test_case = name.split('::')

        test_file = path.split(os.sep)[-1].replace('.py', '')
        return '::'.join(filter(None, [test_file, test_class, test_case]))
    else:
        return name


def put_file(client, source_file, destination_file):
    """
    Copy file to instance using Paramiko client connection.
    """
    try:
        sftp_client = client.open_sftp()
        sftp_client.put(source_file, destination_file)
    except Exception as error:
        raise IpaUtilsException(
            'Error copying file to instance: {0}.'.format(error)
        )
    finally:
        with ignored(Exception):
            sftp_client.close()


@contextmanager
def redirect_output(fileobj):
    """Redirect standard out to file."""
    old = sys.stdout
    sys.stdout = fileobj
    try:
        yield fileobj
    finally:
        sys.stdout = old


@contextmanager
def ssh_config(ssh_user, ssh_private_key_file):
    """Create temporary ssh config file."""
    try:
        ssh_file = NamedTemporaryFile(delete=False, mode='w+')
        ssh_file.write('Host *\n')
        ssh_file.write('    IdentityFile %s\n' % ssh_private_key_file)
        ssh_file.write('    User %s' % ssh_user)
        ssh_file.close()
        yield ssh_file.name
    finally:
        with ignored(OSError):
            os.remove(ssh_file.name)


def update_history_log(history_log,
                       clear=False,
                       description=None,
                       test_log=None):
    """
    Update the history log file with item.
    If clear flag is provided the log file is deleted.
    """
    if not test_log and not clear:
        raise IpaUtilsException(
            'A test log or clear flag must be provided.'
        )

    if clear:
        with ignored(OSError):
            os.remove(history_log)

    else:
        history_dir = os.path.dirname(history_log)
        if not os.path.isdir(history_dir):
            try:
                os.makedirs(history_dir)
            except OSError as error:
                raise IpaUtilsException(
                    'Unable to create directory: %s' % error
                )

        with open(history_log, 'a+') as f:
            # Using append mode creates file if it does not exist
            if description:
                description = '"%s"' % description

            out = '{} {}'.format(
                test_log,
                description or ''
            )
            f.write(out.strip() + '\n')


def get_logger(log_level):
    """
    Return new console logger at provided log level.
    """
    logger = logging.getLogger('img_proof')
    logger.setLevel(log_level)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(logging.Formatter('%(message)s'))

    logger.addHandler(console_handler)
    return logger



# -*- coding: utf-8 -*-

"""SLES distro module and sync points."""

# Copyright (c) 2019 SUSE LLC. All rights reserved.
#
# This file is part of img_proof. img_proof provides an api and command line
# utilities for testing images in the Public Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from img_proof.ipa_distro import Distro
from img_proof.ipa_exceptions import IpaSLESException


class SLES(Distro):
    """SLES distro class."""

    def get_install_cmd(self):
        """Return install package command for SLES."""
        return 'zypper -n --no-gpg-checks in -y'

    def get_refresh_repo_cmd(self):
        """Return refresh repo command for SLES."""
        return 'zypper -n refresh'

    def get_stop_ssh_service_cmd(self):
        """
        Return command to stop SSH service for SLES.

        SSH stop command determined by init system.
        """
        if self.init_system == 'systemd':
            return 'systemctl stop sshd.service'
        elif self.init_system == 'init':
            return 'rcsshd stop'
        else:
            raise IpaSLESException(
                'The init system for SUSE distribution cannot be determined.'
            )

    def get_update_cmd(self):
        """Return command to update SLES instance."""
        return 'zypper -n up --auto-agree-with-licenses ' \
               '--force-resolution --replacefiles'



# -*- coding: utf-8 -*-

"""img_proof constants."""

# Copyright (c) 2019 SUSE LLC. All rights reserved.
#
# This file is part of img_proof. img_proof provides an api and command line
# utilities for testing images in the Public Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os


HOME = os.path.expanduser('~')
NOT_IMPLEMENTED = 'Implement method in child classes.'
SUPPORTED_DISTROS = ('opensuse_leap', 'sles')
SUPPORTED_CLOUDS = ('azure', 'ec2', 'gce', 'ssh', 'oci')

AZURE_DEFAULT_TYPE = 'Standard_B1ms'
AZURE_DEFAULT_USER = 'azureuser'
EC2_DEFAULT_TYPE = 't2.micro'
EC2_DEFAULT_USER = 'ec2-user'
GCE_DEFAULT_TYPE = 'n1-standard-1'
GCE_DEFAULT_USER = 'root'
OCI_DEFAULT_TYPE = 'VM.Standard2.1'
OCI_DEFAULT_USER = 'opc'

EC2_CONFIG_FILE = os.path.join(HOME, '.ec2utils.conf')
IPA_CONFIG_FILE = os.path.join(HOME, '.config', 'img_proof', 'config')

IPA_HISTORY_FILE = os.path.join(HOME, '.config', 'img_proof', '.history')
IPA_RESULTS_PATH = os.path.join(HOME, 'img_proof', 'results')

BASH_SSH_SCRIPT = '''#!/bin/bash
echo {key} >> /home/{user}/.ssh/authorized_keys
'''

SYNC_POINTS = (
    'test_hard_reboot',
    'test_soft_reboot',
    'test_update',
    'test_refresh'
)

TEST_PATHS = (
    os.path.join(os.sep, 'usr', 'share', 'lib', 'img_proof', 'tests', ''),
    os.path.join(HOME, 'img_proof', 'tests', '')
)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""img_proof azure provider unit tests."""

# Copyright (c) 2019 SUSE LLC. All rights reserved.
#
# This file is part of img_proof. img_proof provides an api and command line
# utilities for testing images in the Public Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from img_proof.ipa_azure import AzureCloud
from img_proof.ipa_exceptions import AzureCloudException

from unittest.mock import MagicMock, patch


class TestAzureProvider(object):
    """Test Azure provider class."""

    def setup(self):
        self.client = MagicMock()

    def setup_method(self, method):
        """Set up kwargs dict."""
        self.kwargs = {
            'config': 'tests/data/config',
            'distro_name': 'sles',
            'image_id': 'another:fake:image:id',
            'no_default_test_dirs': True,
            'running_instance_id': 'fakeinstance',
            'service_account_file': 'tests/azure/test-sa.json',
            'ssh_private_key_file': 'tests/data/ida_test',
            'test_dirs': 'tests/data/tests',
            'test_files': ['test_image']
        }

    @patch.object(AzureCloud, '_get_management_client')
    @patch.object(AzureCloud, '_get_ssh_public_key')
    def helper_get_provider(self, mock_get_ssh_pub_key, mock_get_client):
        mock_get_client.return_value = self.client
        return AzureCloud(**self.kwargs)

    def test_azure_exception_required_args(self):
        """Test an exception is raised if required args missing."""
        self.kwargs['ssh_private_key_file'] = None
        msg = 'SSH private key file is required to connect to instance.'

        # Test ssh private key file required
        with pytest.raises(AzureCloudException) as error:
            AzureCloud(**self.kwargs)

        assert str(error.value) == msg

    @patch('img_proof.ipa_azure.get_client_from_auth_file')
    def test_get_management_client(self, mock_get_client):
        client = MagicMock()
        mock_get_client.return_value = client

        client_class = MagicMock()

        provider = self.helper_get_provider()
        result = provider._get_management_client(client_class)

        assert result == client

    @patch('img_proof.ipa_azure.get_client_from_auth_file')
    def test_get_management_client_json_error(self, mock_get_client):
        client_class = MagicMock()
        mock_get_client.side_effect = ValueError(
            'Not valid'
        )

        provider = self.helper_get_provider()

        with pytest.raises(AzureCloudException) as error:
            provider._get_management_client(client_class)

        assert str(error.value) == 'Service account file format is invalid: ' \
            'Not valid.'

    @patch('img_proof.ipa_azure.get_client_from_auth_file')
    def test_get_management_client_key_error(self, mock_get_client):
        client_class = MagicMock()
        mock_get_client.side_effect = KeyError('subscriptionId')

        provider = self.helper_get_provider()

        with pytest.raises(AzureCloudException) as error:
            provider._get_management_client(client_class)

        assert str(error.value) == "Service account file missing key: " \
            "'subscriptionId'."

    @patch('img_proof.ipa_azure.get_client_from_auth_file')
    def test_get_management_client_exception(self, mock_get_client):
        client_class = MagicMock()
        mock_get_client.side_effect = Exception('Not valid')

        provider = self.helper_get_provider()

        with pytest.raises(AzureCloudException) as error:
            provider._get_management_client(client_class)

        assert str(error.value) == 'Unable to create resource management ' \
            'client: Not valid.'

    @patch('img_proof.ipa_azure.ipa_utils.get_public_ssh_key')
    def test_get_ssh_public_key(self, mock_get_pub_key):
        mock_get_pub_key.return_value = b'pub-key'
        provider = self.helper_get_provider()
        key = provider._get_ssh_public_key()

        assert key == 'pub-key'

    def test_azure_is_instance_running(self):
        """Test is instance running method."""
        instance = MagicMock()
        status = MagicMock()
        status.code = 'PowerState'
        status.display_status = 'VM running'
        instance.instance_view.statuses = [status]

        self.client.virtual_machines.get.return_value = instance

        provider = self.helper_get_provider()
        assert provider._is_instance_running()

    @patch.object(AzureCloud, '_wait_on_instance')
    @patch('img_proof.ipa_utils.generate_instance_name')
    def test_azure_launch_instance_exception(
        self, mock_generate_instance_name, mock_wait_on_instance
    ):
        """Test launch instance method."""
        mock_generate_instance_name.return_value = 'azure-test-instance'
        self.client.resource_groups.create_or_update.side_effect = Exception(
            'Cannot create resource group!'
        )
        self.client.resource_groups.delete.side_effect = Exception(
            'Cannot delete resource group!'
        )

        provider = self.helper_get_provider()

        with pytest.raises(AzureCloudException):
            provider._launch_instance()

    @patch.object(AzureCloud, '_wait_on_instance')
    @patch('img_proof.ipa_utils.generate_instance_name')
    def test_azure_launch_instance(
            self, mock_generate_instance_name, mock_wait_on_instance
    ):
        """Test launch instance method."""
        mock_generate_instance_name.return_value = 'azure-test-instance'

        provider = self.helper_get_provider()
        provider._launch_instance()

        assert self.client.network_interfaces.create_or_update.call_count == 1
        assert self.client.public_ip_addresses.create_or_update.call_count == 1
        assert self.client.resource_groups.create_or_update.call_count == 1
        assert self.client.virtual_machines.create_or_update.call_count == 1
        assert self.client.subnets.create_or_update.call_count == 1
        assert self.client.virtual_networks.create_or_update.call_count == 1
        assert provider.running_instance_id == 'azure-test-instance'

    @patch.object(AzureCloud, '_wait_on_instance')
    @patch('img_proof.ipa_utils.generate_instance_name')
    def test_create_storage_profile(
            self, mock_generate_instance_name, mock_wait_on_instance
    ):
        """Test launch instance method."""
        mock_generate_instance_name.return_value = 'azure-test-instance'
        self.kwargs['image_id'] = 'custom-image.vhd'

        image = MagicMock()
        image.name = 'custom-image.vhd'
        image.id = '/id/custom-image.vhd'
        self.client.images.list.return_value = [image]

        provider = self.helper_get_provider()

        provider._process_image_id()
        storage_profile = provider._create_storage_profile()

        assert storage_profile['image_reference']['id'] == \
            '/id/custom-image.vhd'

    def test_process_image_id(self):
        provider = self.helper_get_provider()
        provider._process_image_id()

        assert provider.image_publisher == 'another'
        assert provider.image_offer == 'fake'
        assert provider.image_sku == 'image'
        assert provider.image_version == 'id'

    def test_process_custom_image_id(self):
        provider = self.helper_get_provider()
        provider.image_id = 'custom-image.vhd'

        provider._process_image_id()
        assert provider.image_publisher is None

    def test_set_default_resource_names(self):
        provider = self.helper_get_provider()
        provider._set_default_resource_names()

        assert provider.ip_config_name == 'fakeinstance-ip-config'
        assert provider.nic_name == 'fakeinstance-nic'
        assert provider.public_ip_name == 'fakeinstance-public-ip'

    @patch.object(AzureCloud, '_get_instance')
    def test_azure_set_image_id(self, mock_get_instance):
        """Test set image id method."""
        self.kwargs['image_id'] = None

        image_reference = MagicMock()
        image_reference.publisher = 'another'
        image_reference.offer = 'fake'
        image_reference.sku = 'image'
        image_reference.version = 'id'

        instance = MagicMock()
        instance.storage_profile.image_reference = image_reference

        mock_get_instance.return_value = instance

        provider = self.helper_get_provider()
        provider._set_image_id()

        assert provider.image_id == 'another:fake:image:id'

    @patch('img_proof.ipa_utils.generate_instance_name')
    def test_azure_set_instance_ip_exception(
        self, mock_generate_instance_name
    ):
        """Test set instance ip exception."""
        mock_generate_instance_name.return_value = 'azure-test-instance'
        provider = self.helper_get_provider()

        self.client.public_ip_addresses.get.side_effect = Exception(
            'IP not found'
        )
        self.client.network_interfaces.get.side_effect = Exception(
            'IP not found'
        )

        with pytest.raises(AzureCloudException) as error:
            provider._set_instance_ip()

        assert str(error.value) == \
            'Unable to retrieve instance IP address: IP not found.'

    @patch('img_proof.ipa_utils.generate_instance_name')
    def test_azure_set_instance_ip(self, mock_generate_instance_name):
        """Test set instance ip method."""
        mock_generate_instance_name.return_value = 'azure-test-instance'
        provider = self.helper_get_provider()

        ip_address = MagicMock()
        ip_address.ip_address = '10.0.0.1'
        self.client.public_ip_addresses.get.return_value = ip_address

        provider._set_instance_ip()
        assert provider.instance_ip == '10.0.0.1'

    def test_azure_start_instance(self):
        """Test start instance method."""
        provider = self.helper_get_provider()
        provider.running_instance_id = 'img_proof-test-instance'

        provider._start_instance()
        self.client.virtual_machines.start.assert_called_once_with(
            'img_proof-test-instance', 'img_proof-test-instance'
        )

        # Test exception
        self.client.virtual_machines.start.side_effect = Exception(
            'Instance not found'
        )

        with pytest.raises(AzureCloudException) as error:
            provider._start_instance()

        assert str(error.value) == 'Unable to start instance: ' \
            'Instance not found.'

    def test_azure_stop_instance(self):
        """Test stop instance method."""
        provider = self.helper_get_provider()
        provider.running_instance_id = 'img_proof-test-instance'

        provider._stop_instance()
        self.client.virtual_machines.power_off.assert_called_once_with(
            'img_proof-test-instance', 'img_proof-test-instance'
        )

        # Test exception
        self.client.virtual_machines.power_off.side_effect = Exception(
            'Instance not found'
        )

        with pytest.raises(AzureCloudException) as error:
            provider._stop_instance()

        assert str(error.value) == 'Unable to stop instance: ' \
            'Instance not found.'

    def test_azure_terminate_instance(self):
        """Test terminate instance method."""
        provider = self.helper_get_provider()
        provider.running_instance_id = 'img_proof-test-instance'

        provider._terminate_instance()
        self.client.resource_groups.delete.assert_called_once_with(
            'img_proof-test-instance'
        )

        # Test exception
        self.client.resource_groups.delete.side_effect = Exception(
            'Instance not found'
        )

        with pytest.raises(AzureCloudException) as error:
            provider._terminate_instance()

        assert str(error.value) == 'Unable to terminate resource group: ' \
            'Instance not found.'



# -*- coding: utf-8 -*-

"""Module for testing instances in Azure."""

# Copyright (c) 2020 SUSE LLC. All rights reserved.
#
# This file is part of img_proof. img_proof provides an api and command line
# utilities for testing images in the Public Cloud.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from azure.common.client_factory import get_client_from_auth_file
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.compute import ComputeManagementClient

from img_proof import ipa_utils
from img_proof.ipa_constants import AZURE_DEFAULT_TYPE, AZURE_DEFAULT_USER
from img_proof.ipa_exceptions import AzureCloudException
from img_proof.ipa_cloud import IpaCloud


class AzureCloud(IpaCloud):
    """Class for testing instances in Azure."""

    def __init__(
        self,
        accelerated_networking=False,
        cleanup=None,
        config=None,
        description=None,
        distro_name=None,
        early_exit=None,
        history_log=None,
        image_id=None,
        inject=None,
        instance_type=None,
        log_level=None,
        no_default_test_dirs=False,
        cloud_config=None,
        region=None,
        results_dir=None,
        running_instance_id=None,
        service_account_file=None,
        ssh_private_key_file=None,
        ssh_user=None,
        subnet_id=None,
        test_dirs=None,
        test_files=None,
        timeout=None,
        vnet_name=None,
        vnet_resource_group=None,
        collect_vm_info=None,
        enable_secure_boot=None,
        enable_uefi=None,
        log_callback=None
    ):
        """Initialize Azure Cloud class."""
        super(AzureCloud, self).__init__(
            'azure',
            cleanup,
            config,
            description,
            distro_name,
            early_exit,
            history_log,
            image_id,
            inject,
            instance_type,
            log_level,
            no_default_test_dirs,
            cloud_config,
            region,
            results_dir,
            running_instance_id,
            test_dirs,
            test_files,
            timeout,
            collect_vm_info,
            ssh_private_key_file,
            ssh_user,
            subnet_id,
            enable_secure_boot,
            enable_uefi,
            log_callback
        )

        self.vnet_name = vnet_name or self.ipa_config['vnet_name']
        self.vnet_resource_group = (
                vnet_resource_group or self.ipa_config['vnet_resource_group']
        )

        subnet_args = [
            self.subnet_id, self.vnet_name, self.vnet_resource_group
        ]
        if any(subnet_args) and not all(subnet_args):
            raise AzureCloudException(
                'subnet_id, vnet_resource_group and vnet_name'
                ' are all required to use an existing subnet.'
            )

        self.service_account_file = (
            service_account_file or self.ipa_config['service_account_file']
        )
        if not self.service_account_file:
            raise AzureCloudException(
                'Service account file is required to connect to Azure.'
            )
        else:
            self.service_account_file = os.path.expanduser(
                self.service_account_file
            )

        if not self.ssh_private_key_file:
            raise AzureCloudException(
                'SSH private key file is required to connect to instance.'
            )

        self.accelerated_networking = (
            accelerated_networking or
            self.ipa_config['accelerated_networking']
        )
        self.ssh_user = self.ssh_user or AZURE_DEFAULT_USER
        self.ssh_public_key = self._get_ssh_public_key()

        self.compute = self._get_management_client(ComputeManagementClient)
        self.network = self._get_management_client(NetworkManagementClient)
        self.resource = self._get_management_client(ResourceManagementClient)

        if self.running_instance_id:
            self._set_default_resource_names()

    def _create_network_interface(
        self, ip_config_name, nic_name, public_ip, region,
        resource_group_name, subnet, accelerated_networking=False
    ):
        """
        Create a network interface in the resource group.

        Attach NIC to the subnet and public IP provided.
        """
        nic_config = {
            'location': region,
            'ip_configurations': [{
                'name': ip_config_name,
                'private_ip_allocation_method': 'Dynamic',
                'subnet': {
                    'id': subnet.id
                },
                'public_ip_address': {
                    'id': public_ip.id
                },
            }]
        }

        if accelerated_networking:
            nic_config['enable_accelerated_networking'] = True

        try:
            nic_setup = self.network.network_interfaces.create_or_update(
                resource_group_name, nic_name, nic_config
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to create network interface: {0}.'.format(
                    error
                )
            )

        return nic_setup.result()

    def _create_public_ip(self, public_ip_name, resource_group_name, region):
        """
        Create dynamic public IP address in the resource group.
        """
        public_ip_config = {
            'location': region,
            'public_ip_allocation_method': 'Dynamic'
        }

        try:
            public_ip_setup = \
                self.network.public_ip_addresses.create_or_update(
                    resource_group_name, public_ip_name, public_ip_config
                )
        except Exception as error:
            raise AzureCloudException(
                'Unable to create public IP: {0}.'.format(error)
            )

        return public_ip_setup.result()

    def _create_resource_group(self, region, resource_group_name):
        """
        Create resource group if it does not exist.
        """
        resource_group_config = {'location': region}

        try:
            self.resource.resource_groups.create_or_update(
                resource_group_name, resource_group_config
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to create resource group: {0}.'.format(error)
            )

    def _create_storage_profile(self):
        """
        Create the storage profile for the instance.

        Image reference can be a custom image name or a published urn.
        """
        if self.image_publisher:
            storage_profile = {
                'image_reference': {
                    'publisher': self.image_publisher,
                    'offer': self.image_offer,
                    'sku': self.image_sku,
                    'version': self.image_version
                },
            }
        else:
            for image in self.compute.images.list():
                if image.name == self.image_id:
                    image_id = image.id
                    break
            else:
                raise AzureCloudException(
                    'Image with name {0} not found.'.format(self.image_id)
                )

            storage_profile = {
                'image_reference': {
                    'id': image_id
                }
            }

        return storage_profile

    def _create_subnet(self, resource_group_name, subnet_id, vnet_name):
        """
        Create a subnet in the provided vnet and resource group.
        """
        subnet_config = {'address_prefix': '10.0.0.0/29'}

        try:
            subnet_setup = self.network.subnets.create_or_update(
                resource_group_name, vnet_name, subnet_id, subnet_config
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to create subnet: {0}.'.format(error)
            )

        return subnet_setup.result()

    def _create_virtual_network(self, region, resource_group_name, vnet_name):
        """
        Create a vnet in the given resource group with default address space.
        """
        vnet_config = {
            'location': region,
            'address_space': {
                'address_prefixes': ['10.0.0.0/27']
            }
        }

        try:
            vnet_setup = self.network.virtual_networks.create_or_update(
                resource_group_name, vnet_name, vnet_config
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to create vnet: {0}.'.format(error)
            )

        vnet_setup.wait()

    def _create_vm(self, vm_config):
        """
        Attempt to create or update VM instance based on vm_parameters config.
        """
        try:
            vm_setup = self.compute.virtual_machines.create_or_update(
                self.running_instance_id, self.running_instance_id,
                vm_config
            )
        except Exception as error:
            raise AzureCloudException(
                'An exception occurred creating virtual machine: {0}'.format(
                    error
                )
            )

        vm_setup.wait()

    def _create_vm_config(self, interface):
        """
        Create the VM config dictionary.

        Requires an existing network interface object.
        """
        # Split image ID into it's components.
        self._process_image_id()

        hardware_profile = {
            'vm_size': self.instance_type or AZURE_DEFAULT_TYPE
        }

        network_profile = {
            'network_interfaces': [{
                'id': interface.id,
                'primary': True
            }]
        }

        storage_profile = self._create_storage_profile()

        os_profile = {
            'computer_name': self.running_instance_id,
            'admin_username': self.ssh_user,
            'linux_configuration': {
                'disable_password_authentication': True,
                'ssh': {
                    'public_keys': [{
                        'path': '/home/{0}/.ssh/authorized_keys'.format(
                            self.ssh_user
                        ),
                        'key_data': self.ssh_public_key
                    }]
                }
            }
        }

        vm_config = {
            'location': self.region,
            'os_profile': os_profile,
            'hardware_profile': hardware_profile,
            'storage_profile': storage_profile,
            'network_profile': network_profile
        }

        return vm_config

    def _get_instance(self):
        """
        Return the instance matching the running_instance_id.
        """
        try:
            instance = self.compute.virtual_machines.get(
                self.running_instance_id, self.running_instance_id,
                expand='instanceView'
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to retrieve instance: {0}'.format(error)
            )

        return instance

    def _get_instance_state(self):
        """
        Retrieve state of instance.
        """
        instance = self._get_instance()
        statuses = instance.instance_view.statuses

        for status in statuses:
            if status.code.startswith('PowerState'):
                return status.display_status

    def _get_management_client(self, client_class):
        """
        Return instance of resource management client.
        """
        try:
            client = get_client_from_auth_file(
                client_class, auth_path=self.service_account_file
            )
        except ValueError as error:
            raise AzureCloudException(
                'Service account file format is invalid: {0}.'.format(error)
            )
        except KeyError as error:
            raise AzureCloudException(
                'Service account file missing key: {0}.'.format(error)
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to create resource management client: '
                '{0}.'.format(error)
            )

        return client

    def _get_ssh_public_key(self):
        """
        Generate SSH public key from private key.
        """
        key = ipa_utils.get_public_ssh_key(self.ssh_private_key_file)
        return key.decode()

    def _is_instance_running(self):
        """
        Return True if instance is in running state.
        """
        return self._get_instance_state() == 'VM running'

    def _launch_instance(self):
        """
        Create new test instance in a resource group with the same name.
        """
        self.running_instance_id = ipa_utils.generate_instance_name(
            'azure-img-proof-test'
        )
        self.logger.debug('ID of instance: %s' % self.running_instance_id)
        self._set_default_resource_names()

        try:
            # Try block acts as a transaction. If an exception is raised
            # attempt to cleanup the resource group and all created resources.

            # Create resource group.
            self._create_resource_group(self.region, self.running_instance_id)

            if self.subnet_id:
                # Use existing vnet/subnet.
                subnet = self.network.subnets.get(
                    self.vnet_resource_group, self.vnet_name, self.subnet_id
                )
            else:
                self.subnet_id = ''.join([self.running_instance_id, '-subnet'])
                self.vnet_name = ''.join([self.running_instance_id, '-vnet'])

                # Create new vnet
                self._create_virtual_network(
                    self.region, self.running_instance_id, self.vnet_name
                )

                # Create new subnet in new vnet
                subnet = self._create_subnet(
                    self.running_instance_id, self.subnet_id, self.vnet_name
                )

            # Setup interface and public ip in resource group.
            public_ip = self._create_public_ip(
                self.public_ip_name, self.running_instance_id, self.region
            )
            interface = self._create_network_interface(
                self.ip_config_name, self.nic_name, public_ip, self.region,
                self.running_instance_id, subnet, self.accelerated_networking
            )

            # Get dictionary of VM parameters and create instance.
            vm_config = self._create_vm_config(interface)
            self._create_vm(vm_config)
        except Exception:
            try:
                self._terminate_instance()
            except Exception:
                pass
            raise
        else:
            # Ensure VM is in the running state.
            self._wait_on_instance('VM running', timeout=self.timeout)

    def _process_image_id(self):
        """
        Split image id into component values.

        Example: SUSE:SLES:12-SP3:2018.01.04
                 Publisher:Offer:Sku:Version

        Raises:
            If image_id is not a valid format.
        """
        try:
            image_info = self.image_id.strip().split(':')
            self.image_publisher = image_info[0]
            self.image_offer = image_info[1]
            self.image_sku = image_info[2]
            self.image_version = image_info[3]
        except Exception:
            self.image_publisher = None

    def _set_default_resource_names(self):
        """
        Generate names for resources based on the running_instance_id.
        """
        self.ip_config_name = ''.join([
            self.running_instance_id, '-ip-config'
        ])
        self.nic_name = ''.join([self.running_instance_id, '-nic'])
        self.public_ip_name = ''.join([self.running_instance_id, '-public-ip'])

    def _set_image_id(self):
        """
        If an existing instance is used get image id from deployment.
        """
        instance = self._get_instance()
        image_info = instance.storage_profile.image_reference

        if image_info.publisher:
            self.image_id = ':'.join([
                image_info.publisher, image_info.offer,
                image_info.sku, image_info.version
            ])
        else:
            self.image_id = image_info.id.rsplit('/', maxsplit=1)[1]

    def _set_instance_ip(self):
        """
        Get the IP address based on instance ID.

        If public IP address not found attempt to get private IP.
        """
        try:
            ip_address = self.network.public_ip_addresses.get(
                self.running_instance_id, self.public_ip_name
            ).ip_address
        except Exception:
            try:
                ip_address = self.network.network_interfaces.get(
                    self.running_instance_id, self.nic_name
                ).ip_configurations[0].private_ip_address
            except Exception as error:
                raise AzureCloudException(
                    'Unable to retrieve instance IP address: {0}.'.format(
                        error
                    )
                )

        self.instance_ip = ip_address

    def _start_instance(self):
        """
        Start the instance.
        """
        try:
            vm_start = self.compute.virtual_machines.start(
                self.running_instance_id, self.running_instance_id
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to start instance: {0}.'.format(error)
            )

        vm_start.wait()

    def _stop_instance(self):
        """
        Stop the instance.
        """
        try:
            vm_stop = self.compute.virtual_machines.power_off(
                self.running_instance_id, self.running_instance_id
            )
        except Exception as error:
            raise AzureCloudException(
                'Unable to stop instance: {0}.'.format(error)
            )

        vm_stop.wait()

    def _terminate_instance(self):
        """
        Terminate the resource group and instance.
        """
        try:
            self.resource.resource_groups.delete(self.running_instance_id)
        except Exception as error:
            raise AzureCloudException(
                'Unable to terminate resource group: {0}.'.format(error)
            )

    def get_console_log(self):
        """
        Return console log output if it is available.

        Currently there is no way to get console log from API.
        """
        return ''
